#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import logging

import datetime

_logger = logging.getLogger(__name__)


class ErrorHandler(object):
    """!
    Die Klasse ErrorHandler ist verantwortlich für den Fehlerspeicher der GS1
    """
    ERROR_CODE_UNDEFINED = -2

    def __init__(self, config_manager, communication):
        """!
        Konstruktor
        Initialisierung des Fehler-Mappings und des Fehler-Protokolls
        @param config_manager: Instanz des Config-Managers
        @param communication: Instanz der GUI-Kommunikation
        """
        _logger.debug("start ErrorHandler")
        self._config_manager = config_manager
        self._config_parser = self._config_manager.get_parser()
        self._communication = communication
        try:
            self._error_config_path = self._config_parser.get('path', 'error_config')
            self._error_memory_path = self._config_parser.get('path', 'error_memory')
        except:     # TODO: Exception Type
            pass
        self._load_error_config(self._error_config_path)
        self._error_memory = self._load_error_memory(self._error_memory_path)
        self._reset_all_errors()
        self._goto_error_flag = False
        self._emergency_flag = False

    def _load_error_config(self, error_config_path=None):
        """!
        Lade ein json-File mit der Fehlerbeschreibung/dem Fehlermapping
        @param error_config_path: Pfad zum json-File mit Fehlerbeschreibung/Fehlermapping. Default = None
        """
        if error_config_path is not None:
            try:
                with open(error_config_path, 'r') as data_file:
                    self._error_config = json.load(data_file)
                    _logger.debug(str(type(self._error_config)))
            except:     # TODO: Exception Type
                _logger.error("can not open file" + str(error_config_path))
        else:
            if self._error_config is None:
                self._error_config = json.loads('{"UNDEFINED" : {\
                                                     "text"          : "Undefined",\
                                                     "process_phase" : ["test", "ready", "heat", "hold", "cool"],\
                                                     "error_class"   : 0,\
                                                     "goto_error"    : true,\
                                                     "emergency"     : true,\
                                                     "service"       : true,\
                                                     "max_count"     : 0,\
                                                     "escalate_to"   : "" } }')

    @staticmethod
    def _load_error_memory(error_memory_path=None):
        """!
        Lade ein json-File mit dem Fehlerspeicher
        @param error_memory_path: Pfad zum json-File mit dem Fehlerspeicher. Default = None
        @return Aus der Datei geladener Fehlerspeicher
        """
        error_memory = json.loads('{}')
        try:
            with open(error_memory_path, 'r') as data_file:
                error_memory = json.load(data_file)
                _logger.debug(str(error_memory))
        except:     # TODO: Exception Type
            _logger.error("can not open file" + str(error_memory_path))

        return error_memory

    def set_error(self, error):
        """!
        Trägt einen Fehler in den Fehlerspeicher ein und setzt diesen in den Zustand "aktive".
        Erhöht den Häufigkeitszähler bei bereits eingetragenen Fehlern.
        Eskaliert einen Fehler, wenn dessen maximal erlaubte Häufigkeit erreicht oder überschritten ist.
        @param error: Fehlernummer des einzutragenden Fehlers e.g. F61
        """
        # Prüfen, ob Fehlernummer angelegt ist
        if error in self._error_config.keys():
            # Prüfen, ob Notstopp erforderlich
            if self._get_error_emergency(error):
                self._emergency_flag = True
            # Prüfen, ob in ERROR Zustand gewechselt werden muss
            if self._get_error_goto_error(error):
                self._goto_error_flag = True
            if not self.is_error_active(error):     # Verhindern, dass bereits aktive Fehler erneut eingetragen und geloggt werden
                error_count = self.get_error_count(error) + 1       # Fehlerzähler um 1 erhöhen
                max_count = self._get_error_max_count(error)        # maximal erlaubter Häufigkeitszähler
                if max_count != 0 and error_count >= max_count:     # Muss eskaliert werden?
                    escalate_to = self._get_error_escalate_to(error)
                    self.set_error(escalate_to)
                # Fehler mit Zeitstempel loggen, melden und sichern
                timestamp = "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.datetime.now())
                _logger.error("set_error: {}".format({error: {'count': error_count, 'active': True, 'timestamp': timestamp}}))
                self._communication.write('error=={}'.format(error))
                self._error_memory.update({error: {'count': error_count, 'active': True, 'timestamp': timestamp}})
                if self._error_memory_path is not None and error:
                    try:
                        with open(self._error_memory_path, 'w') as data_file:
                            json.dump(self._error_memory, data_file)
                    except:     # TODO: Exception Type
                        _logger.error("can not open file" + str(self._error_memory_path))
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))

    def reset_error(self, error):
        """!
        Sucht einen Fehler im Fehlerspeicher und setzt dessen Zustand "inaktive".
        @param error: Fehlernummer des zu suchenden Fehlers e.g. F61
        """
        # Prüfen, ob Fehlernummer angelegt ist
        if error in self._error_config.keys():
            # Es können nur aktive (und eingetragene) Fehler zurückgesetzt werden
            if self.is_error_active(error):
                error_count = self.get_error_count(error)
                timestamp = "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.datetime.now())
                _logger.error("reset_error: {}".format({error: {'count': error_count, 'active': False, 'timestamp': timestamp}}))
                self._error_memory.update({error: {'count': error_count, 'active': False, 'timestamp': timestamp}})
                self._communication.write('reseterror=={}'.format(error))
                if self._error_memory_path is not None:
                    try:
                        with open(self._error_memory_path, 'w') as data_file:
                            json.dump(self._error_memory, data_file)
                    except:     # TODO: Exception Type
                        _logger.error("can not open file" + str(self._error_memory_path))
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))

    def _reset_all_errors(self):
        """!
        Setzt alle Fehler im Fehlerspeicher auf den Zustand "inaktive".
        """
        for error in self._error_memory.keys():
            self.reset_error(error)

    def is_error_active(self, error):
        """!
        Sucht einen Fehler im Fehlerspeicher und gibt dessen Zustand zurück.
        @param error: Fehlernummer des zu suchenden Fehlers e.g. F61
        @return True - Fehler ist aktiv, False - Fehler ist nicht aktiv oder nicht eingetragen
        """
        # Prüfen, ob Fehlernummer angelegt ist
        if error in self._error_config.keys():
            error_entry = self._error_memory.get(error)
            # Ist der Fehler eingetragen und aktiv?
            if error_entry is not None and error_entry['active']:
                return True
            else:
                return False
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return False

    def clear_error(self, error, service=False):
        """!
        Löscht einen Fehler aus dem Fehlerspeicher
        @param error: Fehlernummer des zu löschenden Fehlers e.g. F61
        @param service: Muss True sein, damit Service Fehler gelöscht werden können.
        """
        # Prüfen, ob Fehlernummer angelegt ist
        if error in self._error_config.keys():
            if self._get_error_service(error) and service:
                try:
                    del self._error_memory[error]
                except KeyError:
                    pass
                _logger.info("clear_error: {}, {}".format(error, service))
                _logger.debug(str(self._error_memory))
            else:
                _logger.error("can not delete service error: {}".format(str(error)))
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))

    def get_goto_error_flag(self):
        """!
        Prüfe, ob in ERROR Zustand gewechselt werden muss.
        @return True - in Fehlerzustand wechseln, False - nicht in Fehlerzustand wechseln
        """
        return self._goto_error_flag

    def get_emergency_flag(self):
        """!
        Prüfe, ob eine Notabschaltung erforderlich ist
        @return True - Notabschaltung einleiten, False - keine Notabschaltung
        """
        return self._emergency_flag

    def reset_error_flags(self):
        """!
        Setzt alle Infoflags des Errorhandlers zurück.
        """
        self._goto_error_flag = False
        self._emergency_flag = False

    def get_error_count(self, error):
        """!
        Prüfe, ob ein Fehler im Fehlerspeicher steht und gibt den Häufigkeitszähler zurück.
        @param error: Fehlernummer des zu prüfenden Fehlers e.g. F61
        @return Häufigkeitszähler
        """
        error_entry = self._error_memory.get(error)
        if error_entry is None:
            value = 0
        else:
            value = error_entry['count']

        return value

    def _get_error_process_phase(self, error):
        """!
        Gibt die Prozessphase zurück, in der der Fehler auftreten kann.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return Prozessphase ["test", "ready", "heat", "hold", "cool"]
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['process_phase']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return 'all'

    def _get_error_class(self, error):
        """!
        Gibt die Klasse (Schwere) des Fehlers zurück.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return Fehlerklasse - 0: unbekannt 1: leicht, 2: mittel, 3: schwer
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['error_class']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return 0

    def _get_error_goto_error(self, error):
        """!
        Prüft, ob nach Prozessende in den Fehlerzustand gewechselt werden muss.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return True - in Fehlerzustand wechseln, False - nicht in Fehlerzustand wechseln
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['goto_error']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return True

    def _get_error_emergency(self, error):
        """!
        Prüft, ob ein Fehler zur Notabschaltung führt.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return True - Notabschaltung einleiten, False - keine Notabschaltung
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['emergency']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return True

    def _get_error_service(self, error):
        """!
        Prüft, ob ein Fehler durch den Benutzer oder Service zurückgesetzt werden kann.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return True - Service, False - Benutzer
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['service']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return True

    def _get_error_max_count(self, error):
        """!
        Gibt die maximal erlaubte Fehlerhäufigkeit zurück.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return Anzahl der maximal erlaubten Fehler
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['max_count']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return 0

    def _get_error_escalate_to(self, error):
        """!
        Gibt die Nummer des Fehlers zurück, auf den eskaliert werden soll,
        wenn die maximale Fehlerhäufigkeit überschritten wurde.
        @param error: Fehlernummer des umzuwandelnden Fehlers e.g. F61
        @return Fehlernummer für die Fehlereskalation
        """
        if error in self._error_config.keys():
            error_info = self._error_config[error]
            return error_info['escalate_to']
        else:
            _logger.error("error does not exist in error_config: {}".format(str(error)))
            return 0

    def report_error_memory(self):
        """!
        Sendet alle aktiven Fehler an die GUI
        """
        error_message = ''
        for error in self._error_memory.keys():
            # Nur aktive Fehler!
            if self.is_error_active(error):
                if error_message != '':
                    error_message += ','
                error_message += error

        if error_message != '':
            error_message = 'error==' + error_message
            _logger.debug(error_message)
            self._communication.write(error_message)

    def get_error_memory(self):
        """!
        Getter für den Fehlerspeicher
        @return Fehlerspeicher in json-Format
        """
        return self._error_memory

    def clear_error_memory(self, service=False):
        """!
        Löscht den gesamten Fehlerspeicher
        @param service: Wenn True, werden alle Fehler gelöscht, sonst nur Benutzer-löschbare Fehler
        """
        if service:
            self._error_memory = json.loads('{}')
            self.reset_error_flags()
        else:
            for error in self._error_memory.keys():
                self.clear_error(error, service)

        if self._error_memory_path is not None:
            try:
                with open(self._error_memory_path, 'w') as data_file:
                    json.dump(self._error_memory, data_file)
            except:     # TODO: Exception Type
                _logger.error("can not open file" + str(self._error_memory_path))
