#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import logging.handlers
import subprocess
import threading
from threading import Event, Thread
import os
import inspect
import shutil
import time
from datetime import datetime
from platform import system
import ConfigParser
import StringIO


from communication.Communication import Communication
from DigitalAnalogIO import DigitalAnalogIO
from ErrorHandler import ErrorHandler
from HeatFileManager import HeatFileManager
from PowerMonitor import PowerMonitor
from AnalogInput import AnalogInput
from OperatingDataManager import OperatingDataManager
from HeatFileProcessing import HeatFileProcessing
from USBHandler import USBHandler
from driver.RTC import RTC
from controller.ArgonControl import ArgonControl
from controller.DoorControl import DoorControl
from controller.FanControl import FanControl
from controller.VacuumControl import VacuumControl

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = True

ON = True
OFF = False

# Pfad zum SRC_XXXX Ordner
SRC_PATH =  os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# Die letzten vier Zeichen sind die Versionsnummer
SW_VERSION = SRC_PATH[-4:]


class AutoStartTimer(threading._Timer):
    """!
    Von threading.Timer abgeleitete Klasse. Wird zum automatischen Starten von Heizprogrammen verwendet.
    Bekommt zusätzlich die Funktion elapsed und remaining.
    """
    started_at = None

    def start(self):
        """!
        Startet den Timer.
        """
        self.started_at = time.time()
        threading._Timer.start(self)

    def elapsed(self):
        """!
        Gibt die verstrichene Zeit seit dem Start des Timers zurück.
        @return Verstrichene Zeit seit dem Start des Timers in Sekunden
        """
        return time.time() - self.started_at

    def remaining(self):
        """!
        Gibt die verbleibende Zeit bis zur Ausführung der Callback-Funktion zurück.
        @return Verbleibende Zeit in Sekunden
        """
        return self.interval - self.elapsed()


class Gs1Controller(object):
    """!
    GS1Controller.\n
    Steuert den Zustand des Sinterofens.\n
    Startet die Kommunikation mit dem Benutzerinterface.\n
    Verwaltet und startet Heizprozesse.\n
    """
    def __init__(self, config_manager):
        """!
        Initialisierung GS1Controller
        @param config_manager: Instanz des Config Managers
        """
        _logger.debug("GS1Controller init")
        self._name = "Gs1Controller"
        self._config_manager = config_manager
        self._operatingPanel = None
        self.enable_gui_connected = True

        self._readConfigFile()

        self._communication = Communication(self)

        self._error_handler = ErrorHandler(self._config_manager, self._communication)
        self._daio = DigitalAnalogIO(self._error_handler)

        self._hfManager = HeatFileManager(self._config_manager)
        self._usb_handler = USBHandler(self._config_manager, self._communication)

        self._daio.start()
        self._daio.v24_on(ON)
        self._daio.ble_on(ON)
        self._daio.main_switch_green_flash(ON)
        self._operating_data_manager = OperatingDataManager(self._config_manager, self._error_handler)
        self._system_state = "BOOT"
        # Echtzeituhr initialisieren
        self._rtc = RTC()
        self._analog_input = AnalogInput(self._error_handler, self._config_manager)
        # Unterdruck Regelung initialisieren
        self._vacuum_controller = VacuumControl(self._daio, self._analog_input, self._config_manager)
        # Argon Controller initialisieren
        self._argon_controller = ArgonControl(self._daio, self._analog_input, self._config_manager, self._error_handler)
        # Lüfter Controller initialisieren
        self._fan_controller = FanControl(self._daio, self._analog_input)
        self._fan_percent_oven = 0
        # Türsteuerung
        self._doorController = DoorControl(self._daio, self._config_manager, self._error_handler)
        # Heizprofil Controller
        self._heatFileProcessor = HeatFileProcessing(self,
                                                     self._config_manager,
                                                     self._doorController,
                                                     self._analog_input,
                                                     self._error_handler)
        self._power_unit = self._heatFileProcessor.getPowerUnit()
        self._power_monitor = PowerMonitor(self._analog_input, self._config_manager, self._error_handler, self._power_unit)
        self._door_autoclose_timeout = time.time()  # Timer zum automatischen Schließen der Tür
        # self._safety_cycle = False  # Schutz vor durchdrücken z.b. emergency state-> ready state
        # self._ready_button_pressed = False  # Handle short or long pressed
        self._battery_enabled = False

        self._last_gui_state = ""
        self._gui_connected = False
        self._gui_update_started = False
        self._gui_update_finished = False

        self._start_selftest_from_com = False  # this flag is set from communication and start the selftest - only in ready
        # --- thread handling ---
        self._stopped = Event()
        self._stopped.clear()
        self._controlling_thread = None
        self._process_time = 0.100
        self._500ms_cycle_time = time.time() + 0.5
        self._1s_cycle_time = time.time() + 1
        self._3s_cycle_time = time.time() + 3

        self._autoStartTimer = None

        self._sw_version = self._config_manager.get_production_file_value("version", "gs1-sw-version", SW_VERSION)
        self._hw_version = self._config_manager.get_production_file_value("version", "gs1-hw-version")

        # used for logging only
        self._self_test_log_str = '2'  # self test not started
        self._actual_heat_file_name = ""
        self._not_aus_S_sensor = False
        self._not_aus_K_sensor = False
        self._system_update_required = False
        self._heater_fan_self_test_done = False

        # set up data logger
        data_logger_path = self._config_manager.get_parser().get('path', 'gs1_log')
        handler = logging.handlers.RotatingFileHandler(data_logger_path, 'a', 5 * 1024 * 1024, 10)
        formatter = logging.Formatter('')
        handler.setFormatter(formatter)
        self._data_logger = logging.getLogger('data_logger')
        self._data_logger.setLevel(logging.INFO)
        self._data_logger.addHandler(handler)

    def _readConfigFile(self):
        """!
        Funktion veranlasst das Lesen der Ofenkonfiguration aus der Config-Datei.
        """
        # --- config ---
        # serial number oven
        self._oven_serial = self._config_manager.get_production_file_value('system', 'serial-number-oven')
        # oven name
        self._oven_name = self._config_manager.get_production_file_value('system', 'oven-name')
        # Erstinbetriebnahme
        self._initial_startup = int(self._config_manager.get_production_file_value('system', 'initial_startup'))
        # config 20
        self._limit_temperature_cc = float(self._config_manager.get_config_file_value('temp_limits', 'limit-temperature-cc'))
        # config 19
        self._k_limit_temperature = float(self._config_manager.get_config_file_value('temp_limits', 'k-limit-temperature'))

        # Maximal erlaubte Brennraum-Temperatur, wenn die Türe offen ist
        self._max_temperature = float(self._config_manager.get_config_file_value('door', 'max-temperature'))
        self._door_open_close_time = int(self._config_manager.get_config_file_value('door', 'door-closing-time')) / 1000.0
        # Zeit bis zum automatischen Schließen der Tür
        self._door_autoclose_time = float(self._config_manager.get_config_file_value('door', 'door_autoclose_time'))

        # (21) type-forced-ventilation-cc fan or compressed_air
        self._force_cooling_fan = True if self._config_manager.get_config_file_value('heater', 'type-forced-ventilation-cc') == 'fan' else False
        self._force_cooling_compressed_air = True if self._config_manager.get_config_file_value('heater', 'type-forced-ventilation-cc') == 'compressed_air' else False

        # Werte für Gerätekühlung lesen
        self._fan_elo_element_lower_limit = int(self._config_manager.get_config_file_value('devicecooling', 'elo_min'))
        self._fan_elo_element_upper_limit = int(self._config_manager.get_config_file_value('devicecooling', 'elo_max'))

        # Prüfung, ob die elo_min >= elo_max ist und falls ja elo_max = elo_min + 1 setzen,
        # um mögliche Division durch 0 zu verhindern
        if self._fan_elo_element_upper_limit <= self._fan_elo_element_lower_limit:
            self._fan_elo_element_upper_limit = self._fan_elo_element_lower_limit + 1
        self._fan_elo_element_delta_temp = self._fan_elo_element_upper_limit - self._fan_elo_element_lower_limit

        self._fan_k_element_upper_limit = int(self._config_manager.get_config_file_value('devicecooling', 'k-element-max'))
        self._fan_k_element_lower_limit = int(self._config_manager.get_config_file_value('devicecooling', 'k-element-min'))

        # Prüfung, ob die k-element-min >= k-element-max ist und falls ja k-element-max = k-element-min + 1 setzen,
        # um mögliche Division durch 0 zu verhindern
        if self._fan_k_element_upper_limit <= self._fan_k_element_lower_limit:
            self._fan_k_element_upper_limit = self._fan_k_element_lower_limit + 1
        self._fan_k_element_delta_temp = self._fan_k_element_upper_limit - self._fan_k_element_lower_limit

    def start(self):
        """!
        Startet den PowerMonitor, AnalogInput- und Gs1Controller-Thread. Wird nach der Initialisierung des Gs1Controller aufgerufen.
        """
        self._analog_input.start()

        self._stopped.clear()
        self._controlling_thread = Thread(target=self._process, name=self._name)
        self._controlling_thread.setDaemon(True)
        self._controlling_thread.start()
        _logger.debug("GS1Controller start")
        if self._operatingPanel:
            self._operatingPanel.addMessage('GS1Controller started')

    def exit(self):
        self._stopped.set()
        time.sleep(self._process_time)
        self._power_monitor.exit()
        self._heatFileProcessor.exit()
        self._doorController.exit()
        self._analog_input.exit()
        self._operating_data_manager.exit()
        self._daio.exit()
        self._communication.exit()
        time.sleep(self._process_time)
        self._power_monitor = None
        self._power_unit = None
        self._heatFileProcessor = None
        self._doorController = None
        self._analog_input = None
        self._operating_data_manager = None
        self._daio = None
        self._communication = None

    def setOperatingPanel(self, operating_panel):
        self._operatingPanel = operating_panel

    def _load_initial_config(self):
        """!
        In dieser Funktion wird die Produktionsdatendatei ausgelesen.
        Mit den Produktionsdaten werden anschließend verschiedene SW-Module konfiguriert.
        @return True, wenn alle Daten geladen wurden.
        """
        retval = True

        # Netzfrequenzkorrektur der Türlaufzeit vornehmen
        mains_frequency = self._power_monitor.mains_frequency
        if mains_frequency != 0:
            self._doorController.set_mains_frequency(mains_frequency)
        else:
            retval = False

        return retval

    def _process(self):
        """!
        Haupt-Task des GS1Controllers
        100ms Zyklus
        Überwachung USV Timeout
        Kommunikation empfängt und verarbeitet Befehle. Ruft die zugehörenden Callbacks auf.
        Enthält die Gs1Controller Statemachine.
        Führt verschiedene Tasks zu definierten Zeiten aus.
        """
        _logger.debug("{} process running".format(self._name))
        if self._operatingPanel:
            self._operatingPanel.addMessage("{} process running".format(self._name))
        # taster_counter = 0
        self_test_fan_success = 3
        door_self_test_available = True
        door_self_test_done = False
        self_test_door_success = False
        self_test_time_out = time.time() + 10
        door_self_test_time_out = time.time() + (self._door_open_close_time * 4)
        last_process_data = []
        first_time_is_boot = True
        first_time_emergency_stop = True
        first_time_is_heat = True
        first_time_is_ready = True

        while not self._stopped.wait(self._process_time):
            process_start_time = time.time()

            # ------------------------------------------------------------------------------------------------------
            # PowerMonitor.py
            # ------------------------------------------------------------------------------------------------------
            self._do_power_monitoring()

            # ------------------------------------------------------------------------------------------------------
            # Communication.py
            # ------------------------------------------------------------------------------------------------------
            self._do_communication()

            # ------------------------------------------------------------------------------------------------------
            # Statemachine
            # ------------------------------------------------------------------------------------------------------
            ###########################
            # BOOT
            ###########################
            if self._system_state == "BOOT":

                # Produktionsdaten und initiale Konfiguration lesen
                initial_config_done = self._load_initial_config()

                # Ofeninformationen in das Debug-Log schreiben
                _logger.info("language=={}".format(self._operating_data_manager.get_operating_language()))
                _logger.info("unit=={}".format(self._operating_data_manager.get_operating_unit()))
                _logger.info("initial_start-up=={}".format(self._initial_startup))
                _logger.info("name=={}".format(self._oven_name))
                _logger.info("serialno=={}".format(self._oven_serial))
                _logger.info("hwver=={}".format(self._hw_version))
                _logger.info("swver=={}, {}".format(SW_VERSION, self._sw_version))
                _logger.info("ophours=={}".format(self._operating_data_manager.get_operating_hours()))
                _logger.info("country==ROW")
                _logger.info("argon=={}".format(self._argon_controller.getArgonAvailable()))
                _logger.info("maxrate=={}".format(self._hfManager.getMaxHeatRate()))

                if first_time_is_boot and initial_config_done:
                    _logger.debug("gs1 controller boot")
                    if self._operatingPanel:
                        self._operatingPanel.addMessage("gs1 controller boot")
                    self._send_init_message()

                    first_time_is_boot = False

                    if self._rtc.rtcValid():
                        self._error_handler.reset_error("F1010")
                    else:
                        self._error_handler.set_error("F1010")

                    # last_process_data = self._operating_data_manager.load_process_data()

                    if self._initial_startup == 1:
                        self._doorController.door_open(100)
                        self._communication.write("insulationmissing")
                        self._system_state = "DEFECT"
                    # Prüfen, ob SW Update erfolgt ist und erfolgreich war
                    elif self._sw_version == 'verify':
                        if self._verify_sw_update():
                            _logger.info("software update successful")
                            self._system_update_required = False
                            self._communication.write('usb==done')
                            self._sw_version = SW_VERSION
                            self._communication.write("swver=={}".format(self._sw_version))
                            self._system_state = "START_SELFTEST"
                        else:
                            _logger.error("software update failed")
                            self._system_update_required = True
                            self._communication.write('usb==failed')
                            self._door_autoclose_timeout = time.time() + self._door_autoclose_time
                            self._system_state = "DEFECT"
                    else:
                        self._system_update_required = False
                        self._sw_version = SW_VERSION
                        self._system_state = "START_SELFTEST"

            ###########################
            # SELFTEST
            ###########################
            if self._system_state == "START_SELFTEST":
                _logger.debug("start selftest")

                if self._operatingPanel:
                    self._operatingPanel.addMessage("start selftest")
                # Türen-Selbsttest nur starten, wenn Temperatur unter kritischem Level und dann auf Ergebnis warten
                if self._analog_input.temp_tc_one_filtered < self._max_temperature:
                    door_self_test_time_out = time.time() + (self._door_open_close_time * 4)
                    self._doorController.start_door_self_test()
                    door_self_test_available = True
                else:
                    door_self_test_available = False
                    _logger.warning("door selftest not possible => max_temperature")

                self_test_time_out = time.time() + 10
                # Lüftertest starten und auf Ergebnis warten
                self_test_fan_success = 3
                self._fan_controller.setFan1(30)
                self._fan_controller.setFan2(30)
                self._fan_controller.setFan3(30)

                # Heizungstest starten und auf Ergebnis warten
                self._power_unit.startSelftest()

                door_self_test_done = False
                self._heater_fan_self_test_done = False

                self._system_state = "SELFTEST"

            if self._system_state == "SELFTEST":
                # Ende des Lüfter- und Heizungsselbsttests
                if (self_test_time_out < time.time()) and (not self._heater_fan_self_test_done):
                    # Lüfter auf Fehler prüfen
                    if self._fan_controller.fan1_error:
                        self_test_fan_success -= 1
                        _logger.error("SELFTEST_FAN_1_FAIL")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_FAN_1_FAIL')

                    if self._fan_controller.fan2_error:
                        self_test_fan_success -= 1
                        _logger.error("SELFTEST_FAN_2_FAIL")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_FAN_2_FAIL')

                    if self._fan_controller.fan3_error:
                        self_test_fan_success -= 1
                        _logger.error("SELFTEST_FAN_3_FAIL")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_FAN_3_FAIL')

                    if self_test_fan_success == 3:
                        _logger.info("SELFTEST_FAN_OK")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_FAN_OK')
                        self._communication.write('stcooling==pass')
                    else:
                        if self_test_fan_success < 2:
                            self._error_handler.set_error("F50")
                        else:
                            self._error_handler.set_error("F500")
                        self._communication.write('stcooling==fail')

                    # Heizung auf Fehler prüfen
                    if self._power_unit.getSelftestState() == 2:
                        _logger.info("SELFTEST_HEAT_OK")
                        self._communication.write('stheater==pass')
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_HEAT_OK')
                    else:
                        _logger.error("SELFTEST_HEAT_FAIL")
                        self._communication.write('stheater==fail')
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_HEAT_FAIL')

                    self._fan_controller.setFan1(0)
                    self._fan_controller.setFan2(0)
                    self._fan_controller.setFan3(0)

                    self._heater_fan_self_test_done = True

                # Auf Ende des Tür-Selbsttest
                if door_self_test_available:
                    if self._doorController.door_self_test_success:
                        self_test_door_success = True
                        door_self_test_done = True
                        _logger.info("SELFTEST_DOOR_OK")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_DOOR_OK')
                        self._communication.write('stdoor==pass')
                    elif door_self_test_time_out < time.time() or self._doorController.door_error is True:
                        self_test_door_success = False
                        door_self_test_done = True
                        _logger.error("SELFTEST_DOOR_FAIL")
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_DOOR_FAIL')
                        self._communication.write('stdoor==fail')
                elif not door_self_test_done:
                    self_test_door_success = True
                    door_self_test_done = True
                    _logger.warning("SELFTEST_DOOR_NA")
                    if self._operatingPanel:
                        self._operatingPanel.addMessage('SELFTEST_DOOR_NA')
                    self._communication.write('stdoor==na')

                # Gesamtergebnis auswerten
                if door_self_test_done and self._heater_fan_self_test_done:
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time                  # Timeout zum automatischen Schließen der Tür

                    if self_test_door_success and (self_test_fan_success == 3) and (self._power_unit.getSelftestState() == 2):
                        self._self_test_log_str = '1'  # okay
                        _logger.info("SELFTEST_OK")
                        self._communication.write('st==pass')
                        time.sleep(0.15)
                        self._communication.write('st==pass')
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_OK')
                        first_time_is_ready = True
                        self._system_state = "READY"
                    else:
                        self._self_test_log_str = '0'  # not okay
                        _logger.error("SELFTEST_FAIL")
                        self._communication.write('st==fail')
                        time.sleep(0.15)
                        self._communication.write('st==fail')
                        if self._operatingPanel:
                            self._operatingPanel.addMessage('SELFTEST_FAIL')
                        self._system_state = "DEFECT"
                        self._daio.main_switch_green_flash(OFF)
                        self._daio.main_switch_red(ON)

            ###########################
            # DEFECT
            ###########################
            if self._system_state == "DEFECT":
                # Wenn der Timeout zum Tür Schließen abgelaufen ist und
                # wenn die Türe offen ist und
                # wenn die Türisolierung eingesetzt ist
                if (int(self._door_autoclose_timeout - time.time()) < 0 and
                        not self._doorController.is_door_closed() and
                        not self._doorController.is_door_insulation_missing()):
                    _logger.info("Closing door to prevent pollution")
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time      # Timeout neu aufziehen
                    self._doorController.door_close()

                # Selftest initialized from Communication
                if self._start_selftest_from_com:
                    self._start_selftest_from_com = False
                    self._system_state = "START_SELFTEST"

                # TODO: Klärung, ob Taster vorhanden!
                # if self._daio.main_switch_is_set():
                #     taster_counter += 0.05
                #     if taster_counter >= 5:
                #         _logger.debug("more than 5s pressed (Broken) -> Restart")
                #         taster_counter = 0
                #         self.do_reboot()
                # else:
                #     taster_counter = 0

            ###########################
            # READY
            ###########################
            if self._system_state == "READY":
                if first_time_is_ready:
                    self._daio.main_switch_green_flash(OFF)
                    self._daio.main_switch_red_flash(OFF)
                    self._daio.main_switch_green(ON)
                    first_time_is_ready = False

                    if last_process_data != [] and last_process_data[0] != '' and last_process_data[1] != '0':
                        hf_name = last_process_data[0]
                        predry_time = ''
                        stage = int(last_process_data[1]) - 1
                        self.do_start_heat_process(hf_name, predry_time, stage)
                        last_process_data = []
                        self._operating_data_manager.save_process_data()

                # Wenn der Timeout zum Tür Schließen abgelaufen ist und
                # wenn die Türe offen ist und
                # wenn die Türisolierung eingesetzt ist
                #   -> zu machen, damit keine Verschmutzung stattfinden kann.
                if (int(self._door_autoclose_timeout - time.time()) < 0 and
                        not self._doorController.is_door_closed() and
                        not self._doorController.is_door_insulation_missing()):
                    _logger.info("Closing door to prevent pollution")
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time      # Timeout neu aufziehen
                    self._doorController.door_close()

                if self._heatFileProcessor.getHeatingProcessState()[1] != '0':
                    self._system_state = "HEATING"

                # Selftest initialized from Communication
                if self._start_selftest_from_com:
                    self._start_selftest_from_com = False
                    self._system_state = "START_SELFTEST"

                if self._error_handler.get_goto_error_flag():
                    self.do_stop_heat_process()
                    first_time_is_heat = True
                    first_time_is_ready = True
                    self._system_state = "DEFECT"
                    self._operating_data_manager.stop_counter()
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time                  # Timeout zum automatischen Schließen der Tür

                # TODO: Klärung, ob Taster vorhanden!
                # if not self._daio.main_switch_is_set():
                #     # Sicherheit, dass man nicht dauerhaft drücken kann -> Muss im emergency state mindestens einen
                #     # Zyklus bleiben
                #     self._safety_cycle = True
                #
                # if self._safety_cycle:
                #     while self._daio.main_switch_is_set():
                #         # self._ready_button_pressed = True
                #         taster_counter += 0.013
                #         if taster_counter >= 5:
                #             _logger.debug("more than 5s pressed -> Restart")
                #             self._safety_cycle = False
                #             taster_counter = 0
                #             self.do_reboot()
                #         else:
                #             time.sleep(0.01)
                #
                #     # self._ready_button_pressed = False
                #     taster_counter = 0

            ###########################
            # HEAT
            ###########################
            if self._system_state == "HEATING":
                if first_time_is_heat:
                    _logger.debug("gs1 controller heat")
                    self._daio.main_switch_green_flash(ON, 1)
                    self._operating_data_manager.start_counter()
                    first_time_is_heat = False

                # Heizprozess beendet?
                if self._heatFileProcessor.getHeatingProcessState()[1] == '0':
                    first_time_is_heat = True
                    first_time_is_ready = True
                    self._system_state = "READY"
                    self._operating_data_manager.stop_counter()
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time                  # Timeout zum automatischen Schließen der Tür

                    # Ergebnis des Heizvorgangs an GUI senden
                    if self._heatFileProcessor.getHeatingResult():
                        self._communication.write('hpsuccess')
                        time.sleep(0.15)
                        self._communication.write('hpsuccess')
                        time.sleep(0.15)
                        self._communication.write('hpsuccess')
                    else:
                        self._communication.write('hpfailed')
                        time.sleep(0.15)
                        self._communication.write('hpfailed')
                        time.sleep(0.15)
                        self._communication.write('hpfailed')

                if self._error_handler.get_goto_error_flag():
                    self.do_stop_heat_process()
                    first_time_is_heat = True
                    first_time_is_ready = True
                    self._system_state = "DEFECT"
                    self._operating_data_manager.stop_counter()
                    self._door_autoclose_timeout = time.time() + self._door_autoclose_time                  # Timeout zum automatischen Schließen der Tür
                    self._communication.write('hpfailed')

                # TODO: Klärung, ob Taster vorhanden!
                # # main switch 5 sec => not aus
                # if self._daio.main_switch_is_set():
                #     taster_counter += 0.15
                #     if taster_counter >= 5:
                #         _logger.debug("more than 5s pressed (heating)-> to emergency stop")
                #         taster_counter = 0
                #         first_time_is_heat = TRUE
                #         self._system_state = "EMERGENCY_STOP"
                # else:
                #     taster_counter = 0

            ###########################
            # EMERGENCY
            ###########################
            if self._system_state == "EMERGENCY_STOP":
                if first_time_emergency_stop:
                    _logger.critical("gs1 controller emergency")
                    self._heatFileProcessor.abortHeatingProcess()
                    self._daio.main_switch_red_flash(OFF)
                    self._daio.main_switch_green_flash(OFF)
                    self._daio.main_switch_red(ON)
                    self._operating_data_manager.stop_counter()
                    first_time_emergency_stop = False
                    self._fan_percent_oven = 100            # Lüfter auf 100%
                    self._communication.write('hpfailed')
                    self._doorController.door_close()       # Türe zu

                # TODO: Klärung, ob Taster vorhanden!
                # aktuell nicht gewünscht
                # if not self._daio.main_switch_is_set():
                #     # Sicherheit, dass man nicht dauerhaft drücken kann -> Muss im emergency state mindestens einen
                #     # Zyklus bleiben
                #     self._safety_cycle = True
                #
                # main switch to ready
                # if self._daio.main_switch_is_set() and self._safety_cycle:
                #     _logger.debug("pressed (emergency)-> to ready")
                #     self._system_state = "READY"
                #     first_time_emergency_stop = True
                #     self._safety_cycle = False

            # ------------------------------------------------------------------------------------------------------
            # Zyklische Tasks ausführen
            # ------------------------------------------------------------------------------------------------------

            # 100ms Zyklus
            self._100ms_tasks()                                 # 100ms Tasks ausführen

            # 500ms Zyklus
            if self._500ms_cycle_time < time.time():            # 500ms Timer abgelaufen
                self._500ms_cycle_time += 0.5                   # Neuen Timer setzen
                self._500ms_tasks()                             # 500ms Tasks ausführen

            # 1s Zyklus
            if self._1s_cycle_time < time.time():               # 1s Timer abgelaufen
                self._1s_cycle_time += 1                        # Neuen Timer setzen
                self._1s_tasks()                                # 1s Tasks ausführen

            # 3s Zyklus
            if self._3s_cycle_time < time.time():               # 3s Timer abgelaufen
                self._3s_cycle_time += 3                        # Neuen Timer setzen
                self._3s_tasks()                                # 3s Tasks ausführen

            delta_process_time = time.time() - process_start_time
            wait_time = self._process_time - delta_process_time
            if wait_time <= 0:
                if __debug__:
                    _logger.warning("no wait time:{}".format(delta_process_time))

        self._operating_data_manager.stop_counter()

    def _100ms_tasks(self):
        """!
        Enthält alle Funktionen, die im 100ms Zyklus ausgeführt werden.\
        - Notabschaltung bei schwerwiegenden Fehlern\n
        - Temperaturüberwachung\n
        - Lüftersteuerung
        """
        # Prüfen, ob Notabschaltung notwendig ist
        if self._system_state != "EMERGENCY_STOP" and self._error_handler.get_emergency_flag():
            self._system_state = "EMERGENCY_STOP"

        # Fehler Sensor K und Brennkammertemperatur über 300°C
        if (self._error_handler.is_error_active("F400") or self._error_handler.is_error_active("F401")):
            self._daio.main_switch_red_green_flash(ON)
            if self._analog_input.temp_tc_one_filtered > 573:
                self._fan_percent_oven = 100            # Lüfter auf 100%

        # Temperaturüberwachung. Nicht bei Selbsttest, da sonst dessen Ergebnis verfälscht wird.
        if self._heater_fan_self_test_done:
            self._temperature_control()

        # Lüftersteuerung
        self._fan_controller()

    def _500ms_tasks(self):
        """!
        Enthält alle Funktionen, die im 500ms Zyklus ausgeführt werden.\n
        - Kommunikation mit GUI: Initialisierungsbotschaft und zyklischer Ofenzustand.
        """
        if not self._communication.gui_available and self._gui_connected and self.enable_gui_connected:   # Prüfen, ob GUI getrennt wurde
            self._gui_connected = False
            _logger.warning("gui disconnected")
            if self._operatingPanel:
                self._operatingPanel.addMessage('gui disconnected')

        # Wenn GUI verbunden, dann Init Botschaft senden
        if not self._gui_connected or self._last_gui_state == '0':
            self._send_init_message()
        # sonst Alive Botschaft senden
        else:
            self._send_alive_message()
            
            if self._usb_handler.is_gui_update_in_progress():
                # if lineCount < lineCountMax:
                # _logger.info("gui update sollte laufen")
                if not self._usb_handler.gui_update():
                    self._communication.write("usb==failed")

    def _1s_tasks(self):
        """!
        Enthält alle Funktionen, die im 1s Zyklus ausgeführt werden.\n
        - Argonsteuerung\n
        - Unterdruckerzeugung\n
        - USB Handler
        """
        # Argon Controller
        self._argon_controller()
        # Unterdruck Controller
        self._vacuum_controller()

        # Nachsehen, ob ein USB Stick vorhanden ist und ggf. der GUI melden
        # Nur in den Zuständen "READY", "DEFECT" oder "EMERGENCY_STOP"
        if (self._system_state == "READY" or
           self._system_state == "DEFECT" or
           self._system_state == "EMERGENCY_STOP"):
            usbstate = self._usb_handler.monitor()
            if usbstate != '':
                self._communication.write('usb==' + usbstate)

        # Lüfter Fehler überwachen
        fan_error_count = 0

        if not self._force_cooling_fan:
            if self._fan_controller.fan1_error:
                fan_error_count += 1

        if self._fan_controller.fan2_error:
            fan_error_count += 1

        if self._fan_controller.fan3_error:
            fan_error_count += 1

        if not self._power_monitor.on_battery:  # Lüfterprüfung nur bei Netzbetrieb!
            if fan_error_count == 1:
                self._error_handler.set_error("F500")
            elif fan_error_count > 1:
                self._error_handler.set_error("F50")

        # aktuellen zustand der tuer an die gui senden
        if self._doorController.get_door_state() == "CLOSED" or self._doorController.get_door_state() == "CLOSE_GAP":
            self._communication.write("doorclosed")
        elif self._doorController.get_door_state() == "CLOSING":
            self._communication.write("doorclose")
        else:
            self._communication.write("dooropen")

        # Je nach eingetragenem Fehler, muss die Status LED anders leuchten.
        if (self._error_handler.is_error_active("F40") or
            self._error_handler.is_error_active("F41") or
            self._error_handler.is_error_active("F42")):
                self._daio.main_switch_red_flash(ON)
        elif (self._error_handler.is_error_active("F20") or
              self._error_handler.is_error_active("F30") or
              self._error_handler.is_error_active("F32") or
              self._error_handler.is_error_active("F50") or
              self._error_handler.is_error_active("F60") or
              self._error_handler.is_error_active("F61") or
              self._error_handler.is_error_active("F62") or
              self._error_handler.is_error_active("F70") or
              self._error_handler.is_error_active("F80") or
              self._error_handler.is_error_active("F81") or
              self._error_handler.is_error_active("F82") or
              self._error_handler.is_error_active("F8000")):
                self._daio.main_switch_red(ON)
        elif (self._error_handler.is_error_active("F400") or
              self._error_handler.is_error_active("F401") or
              self._error_handler.is_error_active("F500") or
              self._error_handler.is_error_active("F700") or
              self._error_handler.is_error_active("F701")):
                self._daio.main_switch_red_green_flash(ON)

    def _3s_tasks(self):
        """!
        Enthält alle Funktionen, die im 3s Zyklus ausgeführt werden.\n
        - Datenlogger\n
        """
        if self._system_state == "SELFTEST" or self._system_state == "HEATING" or self._system_state == "START_SELFTEST":
            self.data_logging()

    def _temperature_control(self):
        """!
        Liest die Temperaturen von Fühler K und der Elektronik.
        Berechnet anhand der Temperaturen die geforderte Lüfterleistung.
        Fehlerüberwachung des Kühlsystems.
        """
        k_element_temp = self._analog_input.temp_tc_three_filtered

        if k_element_temp < self._fan_k_element_lower_limit:
            fan_percent_k_element = 0
        elif self._fan_k_element_lower_limit <= k_element_temp < self._fan_k_element_upper_limit:
            # linear control
            k_element_temp_delta = k_element_temp - self._fan_k_element_lower_limit
            fan_percent_k_element = int(100.0 / self._fan_k_element_delta_temp * k_element_temp_delta)
        else:
            fan_percent_k_element = 100

        if k_element_temp > self._k_limit_temperature:
            self._error_handler.set_error("F42")

        elo_or_ps_unit_temp = max(self._analog_input.temp_elo_filtered, self._power_unit.getTemperature)
        if elo_or_ps_unit_temp < self._fan_elo_element_lower_limit:
            fan_percent_elo_or_ps_unit = 0
        elif self._fan_elo_element_lower_limit <= elo_or_ps_unit_temp < self._fan_elo_element_upper_limit:
            # linear control
            elo_or_ps_unit_temp_delta = elo_or_ps_unit_temp - self._fan_elo_element_lower_limit
            fan_percent_elo_or_ps_unit = int(100.0 / self._fan_elo_element_delta_temp * elo_or_ps_unit_temp_delta)
        else:
            fan_percent_elo_or_ps_unit = 100

        if elo_or_ps_unit_temp > self._k_limit_temperature:
            self._error_handler.set_error("F42")

        # Entscheiden, welche Temperatur für die Steuerung der Lüfter verwendet wird. Die höhere gewinnt.
        fan_percent = max(self._fan_percent_oven, fan_percent_k_element, fan_percent_elo_or_ps_unit)

        if self._power_monitor.on_battery:  # Lüfterprüfung nur bei Netzbetrieb!
            fan_percent = 0
            
        # Lüfter ansteuern
        if not self._force_cooling_fan:
            self._fan_controller.setFan1(fan_percent)

        self._fan_controller.setFan2(fan_percent)
        self._fan_controller.setFan3(fan_percent)

    def _do_power_monitoring(self):
        """!
        Überwachung des Power Monitors und ggf. Abschaltung nach mehr als 30s Netzausfall.
        """
        if not self._battery_enabled:
            self._daio.enable_battery()
            self._battery_enabled = True

        receive_data = self._power_monitor.from_pm()
        if receive_data is not None:
            _logger.debug("GS1Controller receive: {}".format(receive_data))
            if receive_data == "no-mains-voltage":
                self.do_set_timer("")
            if receive_data == "usv-timeout":
                self._save_system_state()
                _logger.warning("usv-timeout --> shutdown")
                if self._operatingPanel:
                    self._operatingPanel.addMessage('usv-timeout --> shutdown')
                self.do_shutdown()
                _logger.debug("disable battery")
                try:
                    self._daio.disable_battery()
                except AttributeError:
                    pass

    def _do_communication(self):
        """!
        Kommunikation mit der GUI. Sammelt von allen Instanzen Botschaften an die GUI und versendet diese.
        """
        # Botschaften von der Türsteuerung
        message = self._doorController.read()
        if message is not None:
            self._communication.write(message)

        # Botschaften vom Heat File Management
        message = self._hfManager.read()
        if message is not None:
            self._communication.write(message)

        # Botschaften vom Heat File Processing
        message = self._heatFileProcessor.read()
        if message is not None:
            self._communication.write(message)

    def _send_alive_message(self):
        """!
        Erzeugt den Inhalt für die Alive Botschaft und versendet diese an die GUI.
        """
        tempdate = self._rtc.getDate()      # RTC Datum als String im Format YY-MM-DD
        temptime = self._rtc.getTime()      # RTC Uhrzeit als String im Format HH:MM:SS

        current_temperature = self._heatFileProcessor.getCurrentTemperature()
        heating_state = self._heatFileProcessor.getHeatingProcessState()

        # update laeuft gerade
        #_logger.info("ist {} < {} : {}".format(lineCount,
        #                                       lineCountMax,
        #                                       (lineCount < lineCountMax)))

        # Alle Kommandos senden
        self._communication.write("date==20{}{}{}".format(str(tempdate[0:2]), str(tempdate[3:5]), str(tempdate[6:8])))
        self._communication.write("time=={}{}".format(str(temptime[0:2]), str(temptime[3:5])))
        self._communication.write("temp=={:d}".format(int(current_temperature)))                   # Aktuelle Brennraumtemperatur

        if heating_state[1] != '0':
            # Wenn ein Programm läuft, Name, Restlaufzeit [min], Stufe und Heizen/Halten versenden.
            self._communication.write('hpinfo=={};{};{};{}'.format(self._actual_heat_file_name,
                                                                   heating_state[0], heating_state[1], heating_state[2]))
        else:
            # Wenn kein Programm läuft, leere Info senden.
            self._communication.write('hpinfo==')

        # Wenn ein Timer aktiv ist, die verbleibende Zeit bis zum Autostart in [min] ausgeben
        # 29.10.2018 kriek: Wird nicht von GUI verarbeitet und von mv nicht gefordert
        # if self._autoStartTimer is not None:
        #    self._communication.write('timer=={}'.format(str(int(ceil(self._autoStartTimer.remaining()/60)))))

    def _send_init_message(self):
        """!
        Erzeugt den Inhalt für die Init-Botschaft und versendet diese an die GUI.
        """
        tempdate = self._rtc.getDate()      # RTC Datum als String im Format YY-MM-DD
        temptime = self._rtc.getTime()      # RTC Uhrzeit als String im Format HH:MM:SS

        current_temperature = self._heatFileProcessor.getCurrentTemperature()

        # wenn Erstinbetriebnahme, muss das erste Kommando Erstinbetriebnahme sein
        if self._initial_startup == 1:
            self._communication.write("initialstart")
        # Alle Kommandos senden
        self._communication.write("lang=={}".format(self._operating_data_manager.get_operating_language()))
        self._communication.write("unit=={}".format(self._operating_data_manager.get_operating_unit()))
        self._communication.write("date==20{}{}{}".format(str(tempdate[0:2]), str(tempdate[3:5]), str(tempdate[6:8])))
        self._communication.write("time=={}{}".format(str(temptime[0:2]), str(temptime[3:5])))
        self._communication.write("temp=={:d}".format(int(current_temperature)))                   # Aktuelle Brennraumtemperatur
        self._communication.write("name=={}".format(self._oven_name))
        self._communication.write("serialno=={}".format(self._oven_serial))
        self._communication.write("hwver=={}".format(self._hw_version))
        self._communication.write("swver=={}".format(self._sw_version))
        self._communication.write("ophours=={}".format(self._operating_data_manager.get_operating_hours()))
        self._communication.write("country==ROW")
        self._communication.write("argon=={}".format(self._argon_controller.getArgonAvailable()))
        self._communication.write("maxrate=={}".format(self._hfManager.getMaxHeatRate()))
        # self._communication.write("usb=={}".format('unplugged'))  # TODO: self._usb_handler.getUsbState()))
        if self._doorController.is_door_closed():
            self._communication.write("doorclosed")

    def _save_system_state(self):
        """!
        Speichert den aktuellen Zustand der GS1 für den nächsten Systemstart.
        """
        _logger.info("save system state for next startup")
        heating_state = self._heatFileProcessor.getHeatingProcessState()     # Zustand des Prozesses holen
        if heating_state[1] == 'pre':
            heating_state[1] = '1'
        self._operating_data_manager.save_process_data(self._actual_heat_file_name, heating_state[1])

    def _ready_to_heat(self):
        """!
        Prüft, ob alle Bedingungen zum Starten eines Heizprogramms erfüllt sind.\n
        - Systemzustand = READY\n
        - Kein Fehler mit Status "goto_error"-Flag gesetzt
        - Gültiges Heizprofil ist geladen
        - Türisolierung ist eingesetzt
        @return True: Alle Bedingungen zum Starten eines Heizprogramms sind erfüllt
        """

        if (self._system_state != "READY" or                                # Controller Zustand nicht Ready
           self._error_handler.get_goto_error_flag() or                     # goto_error-Fehler eingetragen
           self._actual_heat_file_name == "" or                             # Keine Heizdatei geladen
           self._hfManager.has_material_changed()):                         # Material-/Sinterschalenwechsel erforderlich
                _logger.warning("GS1 is not ready to heat: {}, {}, {}, {}".format(self._system_state,
                                                                                  self._error_handler.get_goto_error_flag(),
                                                                                  self._actual_heat_file_name,
                                                                                  self._hfManager.has_material_changed()))
                retval = False
        else:
            _logger.info("GS1 is ready to heat")
            retval = True

        return retval

    def _verify_sw_update(self):
        """!
        Prüft, ob das SW Update erfolgreich ausgeführt wurde.
        @return True - Update erfolgreich, False - Update nicht erfolgreich
        """
        _logger.info("verify software update")
        # src_ Verzeichnisname holen
        src_path = os.path.basename(os.path.dirname(__file__))
        # Versionsnummer steht an Stelle 4:8
        src_version = src_path[4:]
        # Versionsnummer muss mit dieser SW Version übereinstimmen
        if src_version == SW_VERSION:
            self._config_manager.set_production_file_value("version", "gs1-sw-version", SW_VERSION)
            # Alte SW von GS1 löschen
            old_sw_version = self._config_manager.get_production_file_value("version", "old-sw-version")
            if system() != 'Windows':
                old_src_path = './src_' + old_sw_version
            else:
                old_src_path = '../src_' + old_sw_version
            try:
                shutil.rmtree(old_src_path)
            except:
                _logger.error("cannot delete old src directory {}".format(old_src_path))
            self._config_manager.remove_production_file_value("version", "old-sw-version")
            _logger.info("verification done, new software version is {}".format(SW_VERSION))
            return True
        else:
            _logger.error("software version number {} does not match expected value {}".format(src_version, SW_VERSION))
            return False

    def data_logging(self):
        """!
        Erzeugt die Ausgabe für das Log-File.
        """

        l_str_list = list()

        tempdate = self._rtc.getDate()      # RTC Datum als String im Format YY-MM-DD
        temptime = self._rtc.getTime()      # RTC Uhrzeit als String im Format HH:MM:SS

        log_date_time = "20{}-{}-{} ".format(str(tempdate[0:2]), str(tempdate[3:5]), str(tempdate[6:8]))
        log_date_time += temptime
        l_str_list.append(log_date_time)                                                        # 1: Datums-Zeitstempel
        l_str_list.append(self._oven_serial)                                                    # 2: Ofenkennung
        l_str_list.append(self._actual_heat_file_name)                                          # 3: Heizprogramm Nummer
        l_str_list.append("{:6.0f}".format(self._heatFileProcessor.getHeatingProcessTime()))    # 4: Programmlaufzeit
        l_str_list.append("{:6.0f}".format(self._heatFileProcessor.getEstimatedProcessTime()))  # 5: Programmrestlaufzeit
        l_str_list.append(str(self._operating_data_manager.get_operating_hours()))              # 6: Betriebsstundenzaehler Heizung
        l_str_list.append(str(self._operating_data_manager.get_hf_counter()))                   # 7: Zählerstand durchgeführte Heizprogramme

        l_str_list.append("{:6.1f}".format(self._analog_input.temp_tc_one_filtered))            # 8: Temperaturwert Fühler S (Brennraum)
        l_str_list.append("{:6.1f}".format(self._heatFileProcessor.getSetpointTemperature()))   # 9: Aktuelle Soll Temperatur (Heizprogramm)
        l_str_list.append("{:6.1f}".format(self._analog_input.temp_tc_three_filtered))          # 10: Temperaturwert Fühler K (Ofentechnik)
        l_str_list.append("{:6.1f}".format(self._power_unit.getTemperature))                    # 11: ELO Temperatur Leistungseinheit
        l_str_list.append(self._self_test_log_str)                                              # 12: Selbsttest

        l_str_list.append("1" if self._error_handler.is_error_active("F2010") else "0")         # 13: Versorgungsspannungsüberwachung
        l_str_list.append("{:5.1f}".format(self._power_unit.getMainsVoltage))                   # 14: Netzspannung
        l_str_list.append("{:4.1f}".format(self._power_unit.getMainsFrequency))                 # 15: Netzfrequenz
        l_str_list.append("1" if self._power_unit.getFlagRelayOn else "0")                      # 16: Leistungsausgang (Relais) ON
        l_str_list.append("{:4.1f}".format(self._power_unit.getCurrentEff))                     # 17: Heizstrommessung eff.
        l_str_list.append("{:4.1f}".format(self._power_unit.getCurrentPeak))                    # 18: Heizstrommessung ss.

        l_str_list.append("1" if self._power_unit.getHCDefect else "0")                         # 19: Heizkreis defekt
        l_str_list.append("1" if self._power_unit.getOverCurrent else "0")                      # 20: Überstromschutz (Heizung)
        l_str_list.append("1" if self._power_unit.getOverTemperature else "0")                  # 21: Übertemperaturschutz Leistungseinheit
        l_str_list.append("0")                                                                  # 22: NotAus Taster
        l_str_list.append("0")                                                                  # 23: Platzhalter Systemneustart ausgelöst
        l_str_list.append("1" if self._not_aus_S_sensor else "0")                               # 24: NotAus Übertemperatur Fühler S
        l_str_list.append("1" if self._not_aus_K_sensor else "0")                               # 25: NotAus Übertemperatur Fühler K

        l_str_list.append("1" if self._doorController.door_error else "0")                      # 26: Fehler Türsteuerung
        l_str_list.append("{:3d}".format(self._doorController.actual_percent_door))             # 27: Ist-Stellung Tür
        l_str_list.append("1" if self._fan_controller.fan1_error else "0")                      # 28: Überwachung Lüfter 1
        l_str_list.append("1" if self._fan_controller.fan2_error else "0")                      # 29: Überwachung Lüfter 2
        l_str_list.append("1" if self._fan_controller.fan3_error else "0")                      # 30: Überwachung Lüfter 3
        l_str_list.append("1" if self._analog_input.error_tc_s else "0")                        # 31: Fühlerbruch S
        l_str_list.append("1" if self._analog_input.error_tc_k else "0")                        # 32: Fühlerbruch K

        l_str_list.append("1" if self._gui_connected else "0")                                  # 33: Kommunikation mit Bedieneinheit
        l_str_list.append("1" if self._system_update_required else "0")                         # 34: Softwareupdate
        l_str_list.append("0")  # 35: Logfile kopieren
        l_str_list.append("1" if self._daio.is_fuse_triggered else "0")                         # 36: Elektronische Sicherung
        l_str_list.append("1" if self._power_monitor.on_battery else "0")                       # 37: Batteriebetrieb
        l_str_list.append(str(self._daio.shift_register_bit_string))                            # 38 + 39: Schieberegister DO + DI

        l_str_list.append("{:3d}".format(self._power_unit.getPwrSetValue))                      # 40: Heizleistung Soll
        l_str_list.append("{:3d}".format(self._power_unit.getPwrActualValue))                   # 41: Heizleistung Ist
        l_str_list.append(str(self._power_unit.getPhase))                                       # 42: Leistungsteil Phasensynchronisation
        l_str_list.append("{:5.0f}".format(self._analog_input.pressure_sensor_1))               # 43: Wert Drucksensor 1 in mBar
        l_str_list.append("{:5.0f}".format(self._analog_input.pressure_sensor_2))               # 44: Wert Drucksensor 2 in mBar
        l_str_list.append("{:4.0f}".format(self._analog_input.pressure_sensor_3))               # 45: Wert Drucksensor 3 in mBar
        l_str_list.append("1" if self._argon_controller.getArgonPressureLow() else "0")         # 46: Überwachung Schutzgas Druck niedrig

        self._data_logger.info("; ".join(l_str_list))

        # Anzahl der aktiven Threads in Debug Ausgabe schreiben.
        # _logger.debug("Aktive Threads: {}".format(threading.activeCount()))
        # for th in threading.enumerate():
        #     _logger.debug("{} {}".format(th.getName(), th.is_alive()))


    def enable_gui_connected_detection(self, enable):
        """!
        Schaltet die GUI-Erkennung aus für SW-Update
        @param enable: Erkennung an/aus True/False
        """
        self.enable_gui_connected = enable

    def get_gui_connection_detection(self):
        return self.enable_gui_connected

    ###########################
    # begin cmd2 command interpreter functions
    ###########################

    def set_time(self, new_time):
        """!
        Setzt die RTC Zeit neu
        @param new_time: Zeit als String im Format HHMM
        """
        int_hh = int(new_time[0:2])
        int_mm = int(new_time[2:4])

        # In RTC schreiben
        self._rtc.setTime(int_hh, int_mm, 00)

    def set_date(self, date):
        """!
        Setzt das RTC Datum neu
        @param date: Datum als String im Format YYYYMMDD
        """
        int_yy = int(date[2:4])
        int_mm = int(date[4:6])
        int_dd = int(date[6:8])

        # In RTC schreiben und Systemdatum setzen
        self._rtc.setDate(int_yy, int_mm, int_dd)

    def set_gui_state(self, guistate):
        """!
        Nimmt den GUI Zustand entgegen und sichert diesen in einer Datei
        @param guistate: Zustandsnummer der GUI
        """
        if (guistate != '0') and not self._gui_connected:      # Wenn die GUI auf die Anzeige der Gerätedaten wechselt
            self._gui_connected = True                      # ist die GUI verbunden
            _logger.info("gui connected")
            self.do_load_all_hf_data()                      # ab hier können die Heizdateien übertragen werden
            self._error_handler.report_error_memory()       # Alle aktiven Fehler melden

        self._last_gui_state = guistate                 # da die App selbständig dort hin kommt.

    def get_gui_state(self):
        """!
        Gibt den zuletzt gesicherten GUI Zustand zurück.
        @return Zustandsnummer der GUI
        """
        return self._last_gui_state

    def do_set_timer(self, timer_data):
        """!
        Stellt den Timer für den automatischen Start eines Heizprogrammes.
        @param timer_data: Startinformationen für den Timer im Format YYYYMMDD,HHMM,Name (z.B. 20180717,1520,test123,15).
                           Wird keine Information angegeben, wird der (bereits laufende) Timer gestoppt.
        @return True, wenn der Timer erfolgreich gestellt wurde.
        """
        retval = False
        _logger.info("do_set_timer: {}".format(timer_data))

        if timer_data != '':
            try:
                start_date = timer_data.split(',')[0]
                start_time = timer_data.split(',')[1]
                hf_name = timer_data.split(',')[2]
                predry_time = timer_data.split(',')[3]
            except ValueError:
                _logger.error("error while setting up timer")
                retval = False
            else:
                start = datetime.strptime(start_date + ',' + start_time, '%Y%m%d,%H%M')
                #now = datetime.now()
                #interval = (start - now).total_seconds()
                datestamp = self._rtc.getDate()
                nowDate = '20{}{}{}'.format(datestamp[0:2], datestamp[3:5], datestamp[6:8])
                timestamp = self._rtc.getTime()
                nowTime = timestamp[0:2] + timestamp[3:5] + timestamp[6:8]
                now = datetime.strptime(nowDate + ',' + nowTime, '%Y%m%d,%H%M%S')
                interval = (start - now).total_seconds()

                if interval > 0:
                    self._hfManager.parseHeatFile(hf_name)
                    self._autoStartTimer = AutoStartTimer(interval, self.do_start_heat_process, args=[hf_name, predry_time])     # Timer erzeugen, der den Heizprozess startet.
                    self._autoStartTimer.start()
                    _logger.info("timer enabled, {} seconds left till start of programm '{}'".format(interval, hf_name))
                    retval = True
                else:
                    _logger.error("error while setting up timer")
                    retval = False
        elif self._autoStartTimer is not None:
            _logger.info("timer disabled")
            self._autoStartTimer.cancel()
            self._autoStartTimer = None
            retval = True

        return retval

    def do_safe_hf_data(self, hf_data):
        """!
        Speichert die übergebenen Daten in eine Heizdatei ab. Der Name der Datei steht in den Metainfos.
        @param hf_data: Heizprofil, das gespeichert werden soll.
        @return True, wenn die Datei gespeichert worden sind.
        """
        _logger.debug("do_safe_hf_data: {}".format(hf_data))
        return self._hfManager.safeHeatData(hf_data)

    def do_load_hf_data(self, hf_name, enable_nack=True):
        """!
        Lädt eine Heizdatei und sendet deren Inhalt an die GUI.
        @param hf_name: Name des Heizprofils, das geladen werden soll
        @param enable_nack: Wenn True, wird bei einem Fehler ein 'nack' an die GUI gesendet.
        """
        hf_data = self._hfManager.loadHeatData(hf_name)                       # Inhalt der Heizdatei lesen

        _logger.debug("do_load_hf_data: {}".format(hf_data))

        if hf_data != '':                                                    # Wenn die Datei existiert
            self._communication.write("heatfile=={}".format(hf_data))        # Inhalt der Heizdatei übertragen
            if self._operatingPanel and not enable_nack:
                self._operatingPanel.addMessage("heatfile=={}".format(hf_data))
        elif enable_nack:
            self._communication.write("nack")
            if self._operatingPanel:
                self._operatingPanel.addMessage("nack")

    def do_delete_hf_data(self, hf_name):
        """!
        Löscht eine Heizdatei mit dem angegebenen Namen.
        param hf_name: Name des Heizprofils, das gelöscht werden soll.
        @return True, wenn das Heizprofil erfolgreich gelöscht wurde.
        """
        return self._hfManager.deleteHeatData(hf_name)                       # Heizdatei löschen

    def do_hf_tofavorite(self, hf_name, favorite):
        """!
        Markiert eine Heizdatei als Favorit oder nicht.
        param hf_name: Name des Heizprofils.
        param favorite: True - Als Favorit markieren, False - nicht als Favorit markieren
        @return True, wenn die Einstellung übernommen wurde.
        """
        return self._hfManager.hfToFavorite(hf_name, favorite)

    def do_load_all_hf_data(self):
        """!
        Startet den Thread zur Übertragung aller Heizdateien an die GUI.
        """
        load_all_hf_data_thread = Thread(target=self._load_all_hf_data_target, name="do_load_all_hf_data")
        load_all_hf_data_thread.setDaemon(True)
        load_all_hf_data_thread.start()

        _logger.debug("do_load_all_hf_data start")
        if self._operatingPanel:
            self._operatingPanel.addMessage('do_load_all_hf_data started')

    def _load_all_hf_data_target(self):
        """!
        Sucht alle verfügbaren Heizprofile und sendet deren Inhalt an die GUI.
        """
        hf_names = self._hfManager.listAllHeatFileNames()                   # Alle Heizdateien suchen
        count = 0
        for hf_name in hf_names:                                            # Für alle Namen
            count = count + 1
            time.sleep(0.01)
            self.do_load_hf_data(hf_name, False)                            # Den Inhalt an die GUI übertragen, kein NACK senden

        # doppelt senden um sicher zu gehen, dass alle Daten uebertragen sind
        hf_names = self._hfManager.listAllHeatFileNames()                   # Alle Heizdateien suchen

        for hf_name in hf_names:                                            # Für alle Namen
            time.sleep(0.01)
            self.do_load_hf_data(hf_name, False)                            # Den Inhalt an die GUI übertragen, kein NACK senden

        self._communication.write("hfcount=={}".format(count))

    def do_start_heat_process(self, hf_name='', predry_time='', stage=0):
        """!
        Übergibt Heizprofil an den Ofen und startet den Heizvorgang.
        @param hf_name: Name des Heizprofils.
        @param predry_time: Dauer der Vortrocknungsphase in [min]
        @param stage: Wenn nicht auf 0, startet das Heizprofil in der angegeben Stufe
        @return True, wenn das Heizprofil mit den angegebenen Parametern gestartet wurde.
        """

        _logger.info("do_start_heat_process: {}, {}, {}".format(hf_name, predry_time, stage))

        # Falls ein Timer läuft, diesen Stoppen
        if self._autoStartTimer is not None:
            self._autoStartTimer.cancel()
            self._autoStartTimer = None

        # Wenn eine Datei angegeben wurde, diese laden
        if hf_name != '':
            heating_data = self._hfManager.parseHeatFile(hf_name)
            if heating_data:
                self._actual_heat_file_name = hf_name
            else:
                _logger.error("empty heat profile")
        else:
            heating_data = []
            _logger.error("no heat file name")

        if predry_time != '':
            self.do_set_predry_phase(str(int(predry_time)*60))

        # Wenn bereit, Heizvorgang starten
        _logger.info("do start heat process with file: " + self._actual_heat_file_name)
        if self._ready_to_heat():                           # Prüfen, ob alle Bedingungen zum Starten eines Heizprogramms erfüllt sind
            self._heatFileProcessor.setHeatingData(heating_data)    # In den HF Prozessor geben
            self._heatFileProcessor.startHeatingProcess(stage)      # und den Heizvorgang starten
            self._operating_data_manager.update_hf_counter()
            
            self._fan_percent_oven = 1
            retval = True
        else:
            retval = False

        self._door_autoclose_timeout = time.time() + self._door_autoclose_time

        return retval

    def do_change_hp_data(self, hf_data):
        """!
        Ändert ein Heizprofil während eines laufenden Prozesses..
        @param hf_data: Neues Heizprofil.
        @return True, wenn das Heizprofil mit den angegebenen Parametern geändert wurde.
        """

        _logger.info("do_change_hp_data: {}".format(hf_data))

        # Wenn eine Datei angegeben wurde, diese laden
        if hf_data != '':
            hf_data = hf_data.replace(';', '\n')
            buf = StringIO.StringIO(hf_data)                                # Pseudo Datei mit Heizdaten erzeugen
            heatfileconfig = ConfigParser.ConfigParser()                    # ConfigParser für ini-Dateien
            heatfileconfig.readfp(buf)                                      # Pseudo Datei in ConfigParser laden

            if heatfileconfig.has_section('metainfo'):                      # Datei benötigt metainfo
                try:
                    hf_name = heatfileconfig.get('metainfo', 'name')        # Dateiname lesen
                except ConfigParser.NoOptionError:
                    hf_name = ''

                # Prüfen, ob es sich um die gleiche Datei handelt!
                if hf_name != '' and hf_name == self._actual_heat_file_name:
                    self._hfManager.safeHeatData(hf_data)
                    heating_data = self._hfManager.parseHeatFile(self._actual_heat_file_name)
                    if heating_data:
                        retval = self._heatFileProcessor.setHeatingData(heating_data)  # In den HF Prozessor geben
                    else:
                        _logger.error("empty heat profile")
                        retval = False
                else:
                    _logger.error("can not safe heat data")
                    retval = False
            else:
                _logger.error('heat data not valid, metainfo is missing!')
                retval = False
        else:
            _logger.error("no heat data")
            retval = False

        return retval

    def do_stop_heat_process(self):
        """!
        Beendet einen laufenden Heizvorgang.
        """
        _logger.info("do stop heat process")
        self._heatFileProcessor.abortHeatingProcess()                           # Heizvorgang stoppen
        self._actual_heat_file_name = ""
        self._fan_percent_oven = 0

    def do_set_predry_phase(self, s_duration):
        """!
        (De-)Aktiviert die Vortrocknungsphase.
        @param s_duration: Dauer der Vortrocknungsphase in [s]. Wenn '0', wird keine Vortrocknung gestartet.
        """
        try:
            duration = int(s_duration)                          # als Integer sichern
            self._heatFileProcessor.setPreDryPhase(duration)    # Vortrocknungszeit setzen
            _logger.info("pre dry phase with {}s".format(duration))
        except ValueError:
            self._heatFileProcessor.setPreDryPhase(0)           # Bei Fehler, keine Vortrocknung
            _logger.info("error: pre dry phase disabled")

    def do_conf_mat_changed(self, conf=True):
        """!
        Bestätigt den Wechsel der Sinterschale.
        """
        if conf == True:
            self._hfManager.confirm_material_change()

            if self._actual_heat_file_name != '':
                self.do_start_heat_process(self._actual_heat_file_name)
        else:
            self._hfManager.confirm_material_change(False)

    def do_conf_insulationinstert(self):
        """!
        Bestätigt das Einsetzen der Sinterschale.
        """
        if self._system_state == "DEFECT":
            self._initial_startup = 0
            self._config_manager.remove_production_file_value('system', 'initial_startup')
            self._start_selftest_from_com = True

    def do_move_door(self):
        """!
        Öffnet oder schließt die Tür. Bedingungen:\n
        - Türe ist geschlossen (Endschalter): Türe öffnen\n
        - Türe ist nicht geschlossen: Türe schließen
        @return True, wenn der Befehl ausgeführt werden konnte
        """
        self._heatFileProcessor.setUserInteraction()    # Benutzerinteraktion: Türe

        if self._doorController.get_door_state() == "CLOSED":       # Wenn Türe zu,
            return self.do_door('open')                 # dann öffnen
        elif self._doorController.get_door_state() == "CLOSING":    # Wenn Türe zu geht,
            return self.do_door('open')  # dann öffnen
        elif self._doorController.get_door_state() == "OPEN":  # Wenn Türe offen,
            return self.do_door('close')               # immer schließen
        elif self._doorController.get_door_state() == "OPENING":  # Wenn Türe auf geht,
            return self.do_door('close')                # immer schließen
        elif self._doorController.get_door_state() == "CLOSE_GAP":    # Wenn Türe zu geht,
            return self.do_door('open')  # dann öffnen
        else:
            return self.do_door('close')                # im Zweifel schließen

    def do_door(self, open_close):
        """!
        Öffnen und Schließen der Tür. Erwartet ein String-Argument 'open' oder 'close'.
        @param open_close: open - Öffnet die Türe, close - schließt die Türe
        @return True, wenn der übergebene Befehl korrekt verarbeitet wurde
        """
        _logger.debug("do door: {}".format(open_close))

        retval = False

        args = open_close.split(' ')

        if args[0] == 'close':                                                          # gültiges Argument ist close
            if self._doorController.door_close():
                _logger.info("close door")
                retval = True
        elif args[0] == 'open':                                                         # gültiges Argument ist open
            self._door_autoclose_timeout = time.time() + self._door_autoclose_time      # Timeout zum automatischen Schließen der Tür
            if (len(args) > 1) and (args[1] == 'temp_override'):                        # Ist das Argument temp_override dabei?
                if self._doorController.door_open():                                    # Dann Türe unabhängig von der Brennkammertemperatur öffnen
                    _logger.warning("Temperature override enabled - open door at {} K".format(self._analog_input.temp_tc_one_filtered))
                    retval = True
            elif self._analog_input.temp_tc_one_filtered < self._max_temperature:
                if self._doorController.door_open():
                    _logger.info("open door")
                    retval = True
            else:
                _logger.warning("Can not open door - temperature above safety level")
        else:
            _logger.error('error on door command: {} - use open or close'.format(open_close))

        return retval

    def do_restore_default_settings(self):
        """!
        Werkseinstellung herstellen.
        """
        _logger.warning("restoring default settings")
        self._config_manager.remove_calibration_file_value('powermonitor', 'nominal_mains_voltage')
        self._config_manager.remove_calibration_file_value('powermonitor', 'nominal_mains_frequency')
        self._config_manager.set_production_file_value('system', 'initial_startup', '1')
        self._operating_data_manager.set_operating_language("englishUS")
        self._operating_data_manager.set_operating_unit("c")

        # Fehlerspeicher löschen
        self._error_handler.clear_error_memory(True)
        # RTC zurücksetzen
        self._rtc.resetRtc()

        # Alle Benutzerprogramme löschen
        hf_list = self._hfManager.listAllHeatFileNames(True)
        for hf_name in hf_list:
            self._hfManager.deleteHeatData(hf_name)
    
    def do_reset_error_memory(self):
        """!
        Fehlerspeicher löschen.
        """
        _logger.warning("delete failure memory")
        self._error_handler.clear_error_memory(True)
        self._system_state = "READY"
        self._daio.main_switch_green(ON)

    def do_start_selftest(self):
        """!
        Startet den Ofenselbsttest. Bedingung: Ofen im READY oder DEFECT Zustand.
        @return  True, wenn Selbsttest ausgeführt werden kann
        """
        _logger.debug("do start selftest")
        if (self._system_state == "READY" or self._system_state == "DEFECT"):
            if self._initial_startup == 0:
                self._error_handler.reset_error_flags()
                self._start_selftest_from_com = True
            return True
        else:
            if self._system_state != "SELFTEST":
                self._communication.write('st==fail')
                time.sleep(0.1)
                self._communication.write('st==fail')
            return False

    def do_fc(self, on_off):
        """!
        Schaltet die Zwangskühlung der Brennkammer ein oder aus.
        @param on_off: on - Zwangskühlung an, off - Zwangskühlung aus.
        """
        # Abfrage von == 'on' auf == 'off' geändert -> Sicherer Zustand -> lieber fälschlicher Weise Zwangskühlung ein statt ausgeschaltet
        if on_off == 'off':
            # fan1
            if self._force_cooling_fan:
                self._fan_controller.setFan1(0)
            # air
            elif self._force_cooling_compressed_air:
                self._daio.valve_forced_cooling(False)
            else:
                _logger.error('force_cooling_error')
        elif on_off == 'on':
            # fan1
            if self._force_cooling_fan:
                self._fan_controller.setFan1(100)
            # air
            elif self._force_cooling_compressed_air:
                self._daio.valve_forced_cooling(True)
            else:
                _logger.error('force_cooling_error')
        else:
            # fan1
            if self._force_cooling_fan:
                self._fan_controller.setFan1(100)
            # air
            elif self._force_cooling_compressed_air:
                self._daio.valve_forced_cooling(True)
            else:
                _logger.error('force_cooling_error')
            _logger.error('Wrong argument in command \'fc\': {}'.format(on_off))

    def do_vacuum(self, on_off):
        """!
        Schaltet die Unterdruckerzeugung in der Brennkammer ein oder aus.
        @param on_off: on - Unterdruck an, off - Unterdruck aus.
        """
        if on_off == 'off':
            self._vacuum_controller.enableVacuum()
            _logger.info("Vacuum disabled")
        elif on_off == 'on':
            self._vacuum_controller.disableVacuum()
            _logger.info("Vacuum enabled")
        else:
            _logger.error("Wrong argument in command 'vacuum': {}".format(on_off))
            self._vacuum_controller.disableVacuum()
            _logger.info("Vacuum disabled")

    def do_prot_gas(self, on_off):
        """!
        Schaltet die Argonflutung der Brennkammer ein oder aus.
        @param on_off: on - Argonflutung an, off - Argonflutung aus.
        """
        if on_off == 'off':
            self._argon_controller.disableArgon()
        elif on_off == 'on':
            self._argon_controller.enableArgon()
        else:
            _logger.error('Wrong argument in command \'prot_gas\': {}'.format(on_off))

    def do_vent(self, on_off):
        """!
        Schaltet die Entlüftung der Brennkammer ein oder aus.
        @param on_off: on - Entlüftung an, off - Entlüftung aus.
        """
        # Abfrage von == 'on' auf == 'off' geändert -> Sicherer Zustand -> lieber fälschlicher Weise Entlüftung ein statt ausgeschaltet
        if on_off == 'off':
            self._daio.valve_vent(False)
        elif on_off == 'on':
            self._daio.valve_vent(True)
        else:
            self._daio.valve_vent(True)
            _logger.error('Wrong argument in command \'vent\': {}'.format(on_off))

    def do_res1(self, on_off):
        """!
        Schaltet den Reserveausgang 1 ein oder aus.
        @param on_off: on - Reserveausgang 1 an, off - Reserveausgang 1 aus.
        """
        if on_off == 'off':
            self._daio.v24_out_res1(False)
        elif on_off == 'on':
            self._daio.v24_out_res1(True)
        else:
            _logger.error('Wrong argument in command \'res1\': {}'.format(on_off))

    def do_res2(self, on_off):
        """!
        Schaltet den Reserveausgang 2 ein oder aus.
        @param on_off: on - Reserveausgang 2 an, off - Reserveausgang 2 aus.
        """
        if on_off == 'off':
            self._daio.v24_out_res2(False)
        elif on_off == 'on':
            self._daio.v24_out_res2(True)
        else:
            _logger.error('Wrong argument in command \'res2\': {}'.format(on_off))

    def do_res3(self, on_off):
        """!
        Schaltet den Reserveausgang 3 ein oder aus.
        @param on_off: on - Reserveausgang 3 an, off - Reserveausgang 3 aus.
        """
        if on_off == 'off':
            self._daio.v24_out_res3(False)
        elif on_off == 'on':
            self._daio.v24_out_res3(True)
        else:
            _logger.error('Wrong argument in command \'res3\': {}'.format(on_off))

    def do_res4(self, on_off):
        """!
        Schaltet den Reserveausgang 4 ein oder aus.
        @param on_off: on - Reserveausgang 4 an, off - Reserveausgang 4 aus.
        """
        if on_off == 'off':
            self._daio.rel_res(False)
        elif on_off == 'on':
            self._daio.rel_res(True)
        else:
            _logger.error('Wrong argument in command \'res4\': {}'.format(on_off))

    def do_shutdown(self):
        """!
        Herunterfahren des Betriebssystems mittels 'halt' Befehl und somit Ausschalten des Ofens.
        """
        _logger.warning("Controller shutting down")
        self.data_logging()
        self.exit()
        if system() != 'Windows':
            subprocess.call(["halt"])

    def do_reboot(self):
        """!
        Callback zum Neustarten des Betriebssystems und somit Neustarten des Ofens.
        """
        if self._operatingPanel:
            self._operatingPanel.addMessage("reboot")
        _logger.debug("reboot")
        self.data_logging()
        time.sleep(2)
        self.exit()
        if system() != 'Windows':
            subprocess.call(["reboot"])

    def do_start_gui_update(self):
        """!
        Startet das GUI-Update.\n
        - Lädt das Update auf die GUI\n
        - Veranlasst den Neustart und somit die Übernahme der neuen Software
        @return True, wenn Update erfolgreich.
        """
        self._gui_update_finished = False

        # wenn Start des GUI Update nicht moeglich ist, dann versuche GS1 Update
        if not self._usb_handler.start_gui_update():
            return self.do_update()
        else:
            _logger.info("GUI update started")
            self._gui_update_started = True
            return True

    def do_set_gui_update(self, blockcount):
        return self._usb_handler.resend_gui_update_block(int(blockcount))

    def do_end_gui_update(self):
        _logger.info("GUI update finished")
        self._gui_update_finished = True
        return self._usb_handler._do_end_gui_update()

    def do_update(self):
        """!
        Startet das Software-Update.\n
        - Lädt das Update vom USB Stick auf die GS1\n
        - Veranlasst den Neustart und somit die Übernahme der neuen Software
        @return True, wenn Update erfolgreich.
        """
        if self._gui_update_started:
            _logger.info("GUI update done")
            #self._communication.write("update==finished")
            #self._gui_update_started = False
        _logger.info("GS1 software update")

        # TODO: Implementieren und testen!
        self._system_update_required = True

        if self._system_state == "READY" or self._system_state == "DEFECT":
            retval = self._usb_handler.sw_update(self._sw_version)
        else:
            retval = False

        if retval:
            # Nach Neustart Überprüfung veranlassen
            self._config_manager.set_production_file_value("version", "gs1-sw-version", "verify")
            self._config_manager.set_production_file_value("version", "old-sw-version", SW_VERSION)
            # Bei Erfolg Systemneustart
            self.do_reboot()
        else:
            # sonst 'failed' zurückgeben
            self._communication.write('usb==failed')

            # wenn gui update erfolgreich -> gs1 neu starten
            if self._gui_update_finished:
                AutoStartTimer(2, self.do_reboot).start()

        return retval

    def do_load_programs(self):
        """!
        Lädt Heizprogramme (User-, Hersteller-, Serviceprogramme) von USB Stick auf die GS1.\n
        @return True, wenn das Laden der Heizprogramme erfolgreich durchgeführt wurde.
        """
        _logger.info("Load programs from usb")
        retval = self._usb_handler.load_programs()
        if retval:  # Neue Programme nach dem Laden an die GUI übermitteln
            self.do_load_all_hf_data()
        # bei Erfolg 'done' sonst 'failed' zurückgeben
        self._communication.write('usb=={}'.format('done' if retval else 'failed'))
        return retval

    def do_save_programs(self):
        """!
        Sichert alle User-Heizprogramme von der GS1 auf einen USB Stick.\n
        @return True, wenn das Sichern der Heizprogramme erfolgreich durchgeführt wurde.
        """
        _logger.info("Save user programs to usb")
        retval = self._usb_handler.save_user_programs()
        # bei Erfolg 'done' sonst 'failed' zurückgeben
        self._communication.write('usb=={}'.format('done' if retval else 'failed'))
        return retval

    def do_load_config_files(self):
        """!
        Lädt Konfigurationsdatendateien vom USB Stick auf die GS1.\n
        @return True, wenn das Laden der Konfigurationsdatendateien erfolgreich durchgeführt wurde.
        """
        _logger.info("Load config files from usb")
        retval = self._usb_handler.load_config_files()
        # bei Erfolg 'done' sonst 'failed' zurückgeben
        self._communication.write('usb=={}'.format('done' if retval else 'failed'))
        return retval

    def do_backup(self):
        """!
        Führt ein Backup aller betriebsrelevanten Daten aus. Diese werden auf einen USB Stick gesichert.\n
        @return True, wenn das Sichern der betriebsrelevanten Daten erfolgreich durchgeführt wurde.
        """
        _logger.info("GS1 backup started")
        retval = self._usb_handler.backup()
        # bei Erfolg 'done' sonst 'failed' zurückgeben
        self._communication.write('usb=={}'.format('done' if retval else 'failed'))
        return retval

    def do_safe_operating_language(self, language):
        """!
        Speichert die aktuell in der GUI verwendeten Sprache im operating_data.gs1
        @return True, wenn das Sichern der Sprache erfolgreich war
        """
        return self._operating_data_manager.set_operating_language(language)

    def do_safe_operating_unit(self, unit):
        """!
        Speichert die aktuell in der GUI verwendeten Temperatureinheit im operating_data.gs1
        @return True, wenn das Sichern der Temperatureinheit erfolgreich war
        """
        return self._operating_data_manager.set_operating_unit(unit)
