#!/usr/bin/env python
# -*- coding: utf-8 -*-

import Queue
import logging
import subprocess
from threading import Event, Thread
import time
from platform import system

if system() == 'Windows':
    from Emulator.i2c import i2cget

_logger = logging.getLogger(__name__)


class PowerState(object):
    UNKNOWN = "UNKNOWN"
    BATTERY = "BATTERY"
    AC = "AC"
    USB = "USB"


class TPS65217(object):
    I2CDEV = 0x0
    I2CADDR = 0x24
    CHGCONFIG0 = 0x03
    CHGCONFIG1 = 0x04
    CHGCONFIG2 = 0x05
    CHGCONFIG3 = 0x06
    STATUSREG = 0xA
    ACPWR = 1 << 3
    USBPWR = 1 << 2


class PowerMonitor(object):
    """!
    Überwacht den Zustand des Power Management ICs (PMIC)
    """

    def __init__(self, analog_input, config_manager, error_handler, power_unit):
        """!
        Konstruktor initialisiert die PMIC Überwachung.
        @param analog_input: Instanz der analogen Eingänge
        @param config_manager: Instanz des Config Managers
        @param error_handler: Instanz des Fehlermanagers
        @param power_unit: Instanz der Leistungseinheit
        """
        _logger.debug("PowerMonitor init")
        self._name = "PowerMonitor"
        self._analog_input = analog_input
        self._config_manager = config_manager
        self._error_handler = error_handler
        self._power_unit = power_unit
        # --- thread handling ---
        self._stopped = Event()
        self._powermonitor_thread = None
        self._process_time = 0.5
        self._power_state = PowerState.UNKNOWN
        # --- buffer ---
        self._output_buffer = Queue.Queue()
        self._shutdown_deadline = -1
        self._power_fail = False
        self._mains_voltage = 0.0
        self._mains_voltage_error = False
        self._mains_frequency = 0.0
        self._mains_frequency_error = False

        self._usv_time = int(self._config_manager.get_config_file_value('power_unit', 'usv-time', '30'))
        # Nennwert Netzspannung in Produktionsdaten gesichert?
        # self._mains_voltage = float(self._config_manager.get_calibration_file_value('powermonitor', 'nominal_mains_voltage', '0.0'))
        # Nennwert Netzfrequenz in Produktionsdaten gesichert?
        # self._mains_frequency = float(self._config_manager.get_calibration_file_value('powermonitor', 'nominal_mains_frequency', '0.0'))

        self._start()

    def _start(self):
        """!
        Startet den Power Monitor Thread.
        Konfiguriert das PMIC.
        """
        _logger.debug("PMIC configuration")
        try:
            if system() == 'Windows':
                pass
            else:
                subprocess.check_call(["i2cset", "-y", "-f", str(TPS65217.I2CDEV), str(TPS65217.I2CADDR), str(TPS65217.CHGCONFIG1), "0xB1"])
                subprocess.check_call(["i2cset", "-y", "-f", str(TPS65217.I2CDEV), str(TPS65217.I2CADDR), str(TPS65217.CHGCONFIG2), "0xA0"])
                subprocess.check_call(["i2cset", "-y", "-f", str(TPS65217.I2CDEV), str(TPS65217.I2CADDR), str(TPS65217.CHGCONFIG3), "0xB2"])
        except:     # TODO: Exception Type
            _logger.debug("error during PMIC configuration")

        _logger.debug("PMIC configuration done")

        _logger.debug("PowerMonitor starting")
        self._stopped.clear()
        self._powermonitor_thread = Thread(target=self._process, name=self._name)
        self._powermonitor_thread.setDaemon(True)
        self._powermonitor_thread.start()
        _logger.debug("PowerMonitor started")

    def exit(self):
        """!
        Beendet den Power Monitor Thread
        """
        self._stopped.set()
        time.sleep(self._process_time)
        _logger.debug("PowerMonitor exit()")

    def _process(self):
        """!
        Funktion Prüft zyklisch den Zustand des PMIC.
        """
        startup_delay = 2
        while not self._stopped.wait(self._process_time):
            if startup_delay > 0:
                startup_delay -= 1
            elif self._power_fail is False:
                self.check_power_state()

    def check_power_state(self):
        """!
        Prüft, ob die Stromversorgung durch
          - AC
          - USB
          - oder Batterie
        stattfindet.
        Überwacht den Wert der Netzspannung.
        Überwacht den Wert der Netzfrequenz.
        Startet ggf. den Netzausfall Timer, wenn im Batterie oder USB Betrieb.
        """
        # Netzspannungsüberwachung
        self._check_mains_voltage()

        # Netzfrequenzüberwachung
        self._check_mains_frequency()

        if self._mains_voltage_error or self._mains_frequency_error:
            self._error_handler.set_error("F2010")
        else:
            self._error_handler.reset_error("F2010")

        try:
            if system() == 'Windows':
                power_state_value = i2cget()
            else:
                power_state_value = int(subprocess.check_output(
                    ["i2cget", "-y", "-f", str(TPS65217.I2CDEV), str(TPS65217.I2CADDR), str(TPS65217.STATUSREG)]).strip(),
                                        16)
        except:     # TODO: Exception Type
            return

        if power_state_value & TPS65217.ACPWR == TPS65217.ACPWR:
            new_power_state = PowerState.AC
        elif power_state_value & TPS65217.USBPWR == TPS65217.USBPWR:
            new_power_state = PowerState.USB
        else:
            new_power_state = PowerState.BATTERY

        self._update_power_state(new_power_state)

        if self._shutdown_deadline < 0:
            if self._power_state == PowerState.BATTERY or self._power_state == PowerState.USB:
                self._shutdown_deadline = time.time() + self._usv_time  # check millisecond or second
                self._output_buffer.put("no-mains-voltage")
                _logger.debug(
                    "Power Monitor: GS1 is now on BATTERY or USB and is starting USV timer with {} seconds.".format(
                        self._usv_time))
        else:
            self._check_usv_timeout()

    def _check_mains_voltage(self):
        """!
        Misst und überprüft die Netzspannung. Trägt bei einer Abweichung von +/-10% einen Fehler ein.
        """
        # Nennwert Netzspannung ermittelt?
        if self._mains_voltage > 0.0:
            # Fehler: Netzabweichung um +/-15%
            if not (self._mains_voltage * 0.85 < self._power_unit.getMainsVoltage < self._mains_voltage * 1.15):
                self._mains_voltage_error = True
            else:
                self._mains_voltage_error = False
        else:
            # Nennwert Netzspannung messen und speichern
            self._mains_voltage = self._power_unit.getMainsVoltage
            if self._mains_voltage > 0:
                self._config_manager.set_calibration_file_value('powermonitor', 'nominal_mains_voltage', self._mains_voltage)
                _logger.debug('measured main voltage: {}'.format(self._mains_voltage))

    def _check_mains_frequency(self):
        """!
        Misst und überprüft die Netzfrequenz. Trägt bei einer Abweichung von +/-10% einen Fehler ein.
        """
        # Nennwert Netzfrequenz ermittelt?
        if self._mains_frequency > 0.0:
            # Fehler: Netzabweichung um +/-15%
            if not (self._mains_frequency * 0.85 < self._power_unit.getMainsFrequency < self._mains_frequency * 1.15):
                self._mains_frequency_error = True
            else:
                self._mains_frequency_error = False
        else:
            # Nennwert Netzfrequenz messen und speichern
            self._mains_frequency = self._power_unit.getMainsFrequency
            if self._mains_frequency > 0:
                self._config_manager.set_calibration_file_value('powermonitor', 'nominal_mains_frequency', self._mains_frequency)
                _logger.debug('measured mains frequency: {}'.format(self._mains_frequency))

    def _check_usv_timeout(self):
        """!
        Prüft die Zeit seit Netzausfall (AC) und veranlasst das Abschalten des Ofens.
        """
        current_time = time.time()

        if current_time >= self._shutdown_deadline:
            self._output_buffer.put("usv-timeout")
            _logger.debug(
                "Power Monitor: GS1 was {} seconds on BATTERY or USB and is shutting down now.".format(self._usv_time))
            self._power_fail = True
            if self._analog_input.temp_tc_one_filtered > 1273:          # Temperatur größer 1000°C
                self._error_handler.set_error("F20")

        if self._power_state == PowerState.AC:
            self._shutdown_deadline = -1

    def _update_power_state(self, new_power_state):
        """!
        Prüft die Zeit seit Netzausfall (AC) und veranlasst das Abschalten des Ofens.
        """
        if self._power_state != new_power_state:
            _logger.debug("old power state was " + self._power_state + " new power state is " + new_power_state)
            self._power_state = new_power_state

    def from_pm(self):
        """!
        Gibt den nächsten Eintrag des output Buffers zurück.
        @return Eintrag des Output Buffers
        """
        if not self._output_buffer.empty():
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc
        else:
            return None

    @property
    def on_battery(self):
        """!
        Gibt True zurück, wenn die Stromversorgung über die Batterie erfolgt.
        @return True - Spannungsversorgung erfolgt durch Backup-Batterie
        """
        return True if self._power_state is PowerState.BATTERY else False

    @property
    def mains_frequency(self):
        """!
        Gibt den in der Kalibrierdatendatei hinterlegten Wert der Netzfrequenz zurück.
        @return Wert der in der Kalibrierdatendatei hinterlegten Netzfrequenz
        """
        return self._mains_frequency
